@@ -4,15 +4,13 @@
 Draupnir : Ancestral protein sequence reconstruction using a tree-structured Ornstein-Uhlenbeck variational autoencoder
 =======================
 """
-import torch
-from torch import nn
-import sys
+
 from collections import namedtuple
-#sys.path.append("./draupnir/src/draupnir")
 import draupnir.utils as DraupnirUtils
 from draupnir.models_utils import *
 import pyro
 import pyro.distributions as dist
+
 SamplingOutput = namedtuple("SamplingOutput",["aa_sequences","latent_space","logits","phis","psis","mean_phi","mean_psi","kappa_phi","kappa_psi"])
 class DRAUPNIRModelClass(nn.Module):
     def __init__(self, ModelLoad):
@@ -54,28 +52,29 @@ class DRAUPNIRModelClass(nn.Module):
         self.plate_size = ModelLoad.build_config.plate_subsample_size
         self.plate_unordered = ModelLoad.plate_unordered
         self.one_hot_encoding = ModelLoad.one_hot_encoding
-        if self.batch_size > 1 : # for normal batching and  batch by clade
-            self.n_leaves = self.batch_size
-            self.n_internal = len(self.internal_nodes)
-            self.n_all = self.n_leaves + self.n_internal
-        else:
-            self.n_leaves = len(self.leaves_nodes)
-            self.n_internal = len(self.internal_nodes)
-            self.n_all = self.n_leaves + self.n_internal
+        self.n_leaves_batch = self.batch_size
+        self.n_internal_batch = self.batch_size
+        # if self.batch_size > 1 : # for normal batching and  batch by clade
+        #     self.n_leaves = self.batch_size
+        #     self.n_internal = len(self.internal_nodes)
+        #     self.n_all = self.n_leaves + self.n_internal
+        self.n_leaves = len(self.leaves_nodes)
+        self.n_internal = len(self.internal_nodes)
+        self.n_all = self.n_leaves + self.n_internal
         self.num_layers = 1 #TODO: Remove
         self.h_0_MODEL = nn.Parameter(torch.randn(self.gru_hidden_dim), requires_grad=True).to(self.device)
         if ModelLoad.args.use_cuda:
             self.cuda()
     @abstractmethod
-    def guide(self,family_data, patristic_matrix_sorted,cladistic_matrix,clade_blosum = None):
+    def guide(self, family_data, patristic_matrix, cladistic_matrix, data_blosum, batch_blosum):
         raise NotImplementedError
 
     @abstractmethod
-    def model(self, family_data,patristic_matrix,cladistic_matrix,clade_blosum):
+    def model(self, family_data, patristic_matrix, cladistic_matrix, data_blosum, batch_blosum):
         raise NotImplementedError
 
     @abstractmethod
-    def sample(self, map_estimates, n_samples, family_data_test, patristic_matrix,cladistic_matrix,use_test=True):
+    def sample(self, map_estimates, n_samples, family_data_test, patristic_matrix, cladistic_matrix, use_test=True):
         raise NotImplementedError
     @abstractmethod
     def get_class(self):
@@ -108,6 +107,28 @@ class DRAUPNIRModelClass(nn.Module):
         latent_space = pyro.sample('latent_z', dist.MultivariateNormal(OU_mean, OU_covariance ).to_event(1)) #[z_dim=30,n_nodes] #+ noise[None,:,:]
         latent_space = latent_space.T
         return latent_space
+    def gp_prior_batched(self,patristic_matrix_sorted):
+        "Computes a Gaussian prior over the latent space. The Gaussian prior consists of a Ornstein - Ulenbeck kernel that uses the patristic distances tu build a covariance matrix"
+        # Highlight; OU kernel parameters #TODO: Add noise to OU parameters to avoid error in cholesky decomposition
+        alpha = pyro.sample("alpha", dist.HalfNormal(1).expand_by([3, ]).to_event(1))
+        #alpha = pyro.sample("alpha", dist.Weibull(1,5).expand_by([3, ]).to_event(1))
+        sigma_f = pyro.sample("sigma_f", dist.HalfNormal(alpha[0]).expand_by([self.z_dim, ]).to_event(1))  # rate of mean reversion/selection strength---> signal variance #removed .to_event(1)...
+        sigma_n = pyro.sample("sigma_n", dist.HalfNormal(alpha[1]).expand_by([self.z_dim, ]).to_event(1))  # Gaussian noise
+        lambd = pyro.sample("lambd", dist.HalfNormal(alpha[2]).expand_by([self.z_dim, ]).to_event(1))  # characteristic length-scale
+        # Highlight: Sample the latent space from MultivariateNormal with GP prior on covariance
+        patristic_matrix = patristic_matrix_sorted[1:, 1:]
+        OU_covariance = OUKernel_Fast(sigma_f, sigma_n, lambd).forward(patristic_matrix)
+        OU_mean = torch.zeros((patristic_matrix.shape[0],)).unsqueeze(0)
+        assert OU_covariance.shape == (self.z_dim, self.n_leaves_batch, self.n_leaves_batch)
+        assert OU_mean.shape == (1,self.n_leaves_batch)
+        #noise = 1e-15 + torch.eye(OU_covariance.shape[1])
+        #https://github.com/pyro-ppl/pyro/issues/702
+        #https://forum.pyro.ai/t/runtimeerror-during-cholesky-decomposition/1216/2---> fix runtime error with choleky decomposition
+        #https://forum.pyro.ai/t/using-constraints-within-an-nn-module/486
+        #OU_covariance = transform_to(constraints.lower_cholesky)(OU_covariance) #check that this does not affect performance
+        latent_space = pyro.sample('latent_z', dist.MultivariateNormal(OU_mean, OU_covariance ).to_event(1)) #[z_dim=30,n_nodes] #+ noise[None,:,:]
+        latent_space = latent_space.T
+        return latent_space
 
     def map_sampling(self,map_estimates,patristic_matrix_full):
         "Use map sampling for leaves prediction/testing, when internal nodes are not available"
@@ -115,7 +136,6 @@ class DRAUPNIRModelClass(nn.Module):
         latent_space_internal = map_estimates["latent_z"][:, test_indexes].T
         assert latent_space_internal.shape == (self.n_internal, self.z_dim)
         return latent_space_internal
-
     def conditional_sampling(self,map_estimates, patristic_matrix):
             """Conditional sampling the internal nodes given the leaves from a Multivariate Normal according to page 698 at Pattern Recognition and ML (Bishop)
             :param map_estimates: dictionary conatining the MAP estimates for the OU process parameters
@@ -208,7 +228,51 @@ class DRAUPNIRModelClass(nn.Module):
             latent_space = latent_space.T
             assert latent_space.shape == (self.n_internal, self.z_dim)
             return OU_mean.squeeze(-1).T
+    def conditional_sampling_batch(self,map_estimates, patristic_matrix):
+            """Conditional sampling from Multivariate Normal according to page 698 at Pattern Recognition and ML (Bishop)"""
+            sigma_f = map_estimates["sigma_f"] + 1e-6
+            sigma_n = map_estimates["sigma_n"] + 1e-6
+            lambd = map_estimates["lambd"] + 1e-6
+            internal_indexes = (patristic_matrix[1:, 0][..., None] == self.internal_nodes_batch).any(-1)
+            #n_internal = family_data_test.shape[0]
+            # Highlight: Sample the ancestors conditiones on the leaves (by using the full patristic matrix). See Page 689 at Patter Recongnition and Ml (Bishop)
+            # Highlight: Formula is: p(xa|xb) = N (x|µa|b, Λ−1aa ) , a = test/internal; b= train/leaves
+            patristic_matrix_batch = patristic_matrix[1:, 1:]
+            assert patristic_matrix_batch.shape == (self.n_leaves_internal_batch, self.n_leaves_internal_batch), "Here we are using a slice of the patristic matrix with size n_leaves_batch = batch_size!"
+            OU = OUKernel_Fast(sigma_f, sigma_n, lambd)
+            OU_covariance_full = OU.forward(patristic_matrix_batch)
+            # Highlight: Calculate the inverse of the covariance matrix Λ ≡ Σ−1
+            Inverse_full = torch.linalg.inv(OU_covariance_full)  # [z_dim,n_test+n_train,n_test+n_train]
+            assert Inverse_full.shape == (self.z_dim, self.n_leaves_internal_batch, self.n_leaves_internal_batch)
+            # Highlight: B.49 Λ−1aa
+            Inverse_internal = Inverse_full[:, internal_indexes, :]
+            Inverse_internal = Inverse_internal[:, :, internal_indexes]  # [z_dim,n_test,n_test]
+            assert Inverse_internal.shape == (self.z_dim, self.n_internal_batch, self.n_internal_batch)
+            # Highlight: Conditional mean Mean ---->B-50:  µa|b = µa − Λ−1aa Λab(xb − µb)
+            # Highlight: µa
+            OU_mean_internal = torch.zeros((self.n_internal_batch,))  # [n_internal,]
+            # Highlight: Λab
+            Inverse_internal_leaves = Inverse_full[:,internal_indexes]  # [z_dim,n_test,n_test+n_train]---> [z_dim,n_train,]
+            Inverse_internal_leaves = Inverse_internal_leaves[:, :, ~internal_indexes]  # [z_dim,n_test,n_train]
+            assert Inverse_internal_leaves.shape == (self.z_dim, self.n_internal_batch, self.n_leaves)
+            # Highlight: xb
+            xb = map_estimates["latent_z"]  # [z_dim,n_train]
+            # if self.leaves_testing:
+            #     leaves_indexes = (patristic_matrix[1:, 0][..., None] == self.leaves_nodes).any(-1) #only the indexes of the training leaves
+            #     xb = xb[:,leaves_indexes]
+            # Highlight:µb
+            OU_mean_leaves = torch.zeros((self.n_leaves,))
+            # Highlight:µa|b---> Splitted Equation  B-50
+            part1 = torch.matmul(torch.linalg.inv(Inverse_internal), Inverse_internal_leaves)  # [z_dim,n_test,n_train]
+            part2 = xb - OU_mean_leaves[None, :]  # [z_dim,n_train]
+            OU_mean = OU_mean_internal[None, :, None] - torch.matmul(part1, part2[:, :,None])  # [:,n_test,:] - [z_dim,n_test,None]
+            assert OU_mean.squeeze(-1).shape == (self.z_dim, self.n_internal_batch)
+            latent_space = dist.MultivariateNormal(OU_mean.squeeze(-1), torch.linalg.inv(Inverse_internal)).to_event(1).sample()
+            #latent_space = dist.MultivariateNormal(OU_mean.squeeze(-1), torch.cholesky_inverse(Inverse_internal) + 1e-6).to_event(1).sample()
 
+            latent_space = latent_space.T
+            assert latent_space.shape == (self.n_internal_batch, self.z_dim)
+            return latent_space
 
 class DRAUPNIRModel_classic(DRAUPNIRModelClass):
     """Implements the ordinary version of Draupnir. It receives as an input the entire leaves dataset, uses a GRU as the mapping function and blosum embeddings"""
@@ -218,7 +282,7 @@ class DRAUPNIRModel_classic(DRAUPNIRModelClass):
         self.num_layers = 1
         self.decoder = RNNDecoder_Tiling(self.align_seq_len, self.aa_probs, self.gru_hidden_dim, self.z_dim, self.rnn_input_size,self.kappa_addition,self.num_layers,self.pretrained_params)
         self.embed = EmbedComplex(self.aa_probs,self.embedding_dim, self.pretrained_params)
-    def model(self, family_data, patristic_matrix_sorted,cladistic_matrix,clade_blosum = None):
+    def model(self, family_data, patristic_matrix_sorted,cladistic_matrix,data_blosum,batch_blosum=None):
         aminoacid_sequences = family_data[:, 2:, 0]
         #batch_nodes = family_data[:, 0, 1]
         #batch_indexes = (patristic_matrix_sorted[1:, 0][..., None] == batch_nodes).any(-1)
@@ -292,7 +356,7 @@ class DRAUPNIRModel_classic_no_blosum(DRAUPNIRModelClass):
         DRAUPNIRModelClass.__init__(self,ModelLoad)
         self.rnn_input_size = self.z_dim
         self.decoder = RNNDecoder_Tiling(self.align_seq_len, self.aa_probs, self.gru_hidden_dim, self.z_dim, self.rnn_input_size,self.kappa_addition,self.num_layers,self.pretrained_params)
-    def model(self, family_data, patristic_matrix_sorted,cladistic_matrix,clade_blosum = None):
+    def model(self, family_data, patristic_matrix_sorted,cladistic_matrix,data_blosum,batch_blosum = None):
         aminoacid_sequences = family_data[:, 2:, 0]
         batch_nodes = family_data[:, 0, 1]
         # Highlight: Register GRU module
@@ -364,7 +428,7 @@ class DRAUPNIRModel_classic_plating(DRAUPNIRModelClass):
             self.model = self.model_unordered
         else:
             self.model = self.model_ordered
-    def model_ordered(self, family_data, patristic_matrix_sorted,cladistic_matrix,clade_blosum = None):
+    def model_ordered(self, family_data, patristic_matrix_sorted,cladistic_matrix,data_blosum,batch_blosum = None):
         aminoacid_sequences = family_data[:, 2:, 0]
         batch_nodes = family_data[:, 0, 1]
         #batch_indexes = (patristic_matrix_sorted[1:, 0][..., None] == batch_nodes).any(-1)
@@ -389,7 +453,7 @@ class DRAUPNIRModel_classic_plating(DRAUPNIRModelClass):
                     hidden=decoder_hidden[:,indx])
                 pyro.sample("aa_sequences", dist.Categorical(logits=logits),obs=aminoacid_sequences[indx])  # aa_seq = [n_nodes,align_seq_len]
 
-    def model_unordered(self, family_data, patristic_matrix_sorted,cladistic_matrix,clade_blosum = None):
+    def model_unordered(self, family_data, patristic_matrix_sorted,cladistic_matrix,data_blosum,batch_blosum = None):
         aminoacid_sequences = family_data[:, 2:, 0]
         batch_nodes = family_data[:, 0, 1]
         #batch_indexes = (patristic_matrix_sorted[1:, 0][..., None] == batch_nodes).any(-1)
@@ -415,8 +479,6 @@ class DRAUPNIRModel_classic_plating(DRAUPNIRModelClass):
                     hidden=decoder_hidden[:,indx])
                 pyro.sample("aa_sequences", dist.Categorical(logits=logits),obs=aminoacid_sequences[indx])  # aa_seq = [n_nodes,align_seq_len]
 
-
-
     def sample(self, map_estimates, n_samples, family_data_test, patristic_matrix,cladistic_matrix,use_argmax=False,use_test=True,use_test2=False):
         if use_test or use_test2:
             assert patristic_matrix[1:,1:].shape == (self.n_all,self.n_all)
@@ -456,24 +518,16 @@ class DRAUPNIRModel_classic_plating(DRAUPNIRModelClass):
 
         return sampling_out
 
-class DRAUPNIRModel_plating(DRAUPNIRModelClass):
-    """Implements the plated version of Draupnir.
-     a) It receives as an input the entire leaves dataset and patristic distance matrix to perform full inference over the latent space
-     b) Plates or subsamples the sequences when mapping them to the observations, SPLITS the blosum embeddings accordingly.
-     c) uses a GRU as the mapping function from z to aa"""
+class DRAUPNIRModel_batching(DRAUPNIRModelClass):
+    "Blosum weighted average embedding. Training on leaves, testing on internal nodes, no batching"
     def __init__(self,ModelLoad):
         DRAUPNIRModelClass.__init__(self,ModelLoad)
         self.rnn_input_size = self.z_dim + self.aa_probs
-
-        #self.decoder_attention = RNNAttentionDecoder(self.n_leaves, self.align_seq_len, self.aa_probs, self.gru_hidden_dim,self.rnn_input_size,self.embedding_dim, self.z_dim, self.kappa_addition)
-        self.num_layers = 1
         self.decoder = RNNDecoder_Tiling(self.align_seq_len, self.aa_probs, self.gru_hidden_dim, self.z_dim, self.rnn_input_size,self.kappa_addition,self.num_layers,self.pretrained_params)
         self.embed = EmbedComplex(self.aa_probs,self.embedding_dim, self.pretrained_params)
-        self.n_splits =  int(self.n_leaves/self.plate_size)
-        self.splitted_leaves_indexes = list(torch.tensor_split(torch.arange(self.n_leaves), int(self.n_leaves/self.plate_size))*self.num_epochs)
-
-
-    def model(self, family_data, patristic_matrix_sorted,cladistic_matrix,subsample_split):
+        self.internal_nodes_batch = None
+        self.n_leaves_internal_batch = None
+    def model(self, family_data, patristic_matrix_sorted,cladistic_matrix,data_blosum,batch_blosum):
         aminoacid_sequences = family_data[:, 2:, 0]
         batch_nodes = family_data[:, 0, 1]
         batch_indexes = (patristic_matrix_sorted[1:, 0][..., None] == batch_nodes).any(-1)
@@ -481,58 +535,52 @@ class DRAUPNIRModel_plating(DRAUPNIRModelClass):
         pyro.module("embeddings",self.embed)
         pyro.module("decoder", self.decoder)
         # Highlight: GP prior over the latent space
-        latent_space = self.gp_prior(patristic_matrix_sorted)
+        latent_space = self.gp_prior_batched(patristic_matrix_sorted)
         # Highlight: MAP the latent space to logits using the Decoder from a Seq2seq model with/without attention
         latent_space = latent_space.repeat(1,self.align_seq_len).reshape(latent_space.shape[0],self.align_seq_len,self.z_dim) #[n_nodes,max_seq,z_dim]
-        # blosum = self.blosum_weighted.repeat(latent_space.shape[0],1).reshape(latent_space.shape[0],self.align_seq_len,self.aa_probs) #[n_nodes,max_seq,21]
-        # blosum = self.embed(blosum) #TODO: Introduce a noise variable to be able to deal with more random mutations?
-        # latent_space = torch.cat((latent_space,blosum),dim=2) #[n_nodes,align_seq_len,z_dim + 21]
-        decoder_hidden = self.h_0_MODEL.expand(self.decoder.num_layers * 2,
-                                               latent_space.shape[0],
-                                               self.gru_hidden_dim).contiguous()  # bidirectional #TODO: same hidden state sliced or new hidden states for every subsample
-        #next(iter(self.splitted_leaves_indexes)))
-
-        with pyro.plate("plate_len", aminoacid_sequences.shape[1], dim=-1):
-            #with pyro.plate("plate_seq",aminoacid_sequences.shape[0],dim=-2,subsample_size=self.plate_size) as indx:#Highlight: Random subsampling
-            with pyro.plate("plate_seq", aminoacid_sequences.shape[0], dim=-2,subsample= self.splitted_leaves_indexes.pop(0))as indx:  # Highlight: Ordered subsampling
-                latent_space_subsample = latent_space[indx]
-                aa_frequencies = DraupnirUtils.calculate_aa_frequencies_torch(aminoacid_sequences[indx],self.aa_probs) #TODO, merge in class function
-                blosum_max, blosum_weighted, variable_score = DraupnirUtils.process_blosum(self.blosum, aa_frequencies,self.align_seq_len,self.aa_probs)
-                blosum_embedding = blosum_weighted.repeat(latent_space_subsample.shape[0], 1).reshape(latent_space_subsample.shape[0],
-                                                                                       self.align_seq_len,
-                                                                                       self.aa_probs)  # [n_nodes,max_seq,21]
-                blosum_embedding = self.embed(blosum_embedding)
-                latent_space_subsample = torch.cat((latent_space_subsample, blosum_embedding), dim=2)  # [n_nodes,align_seq_len,z_dim + 21]
-                logits = self.decoder.forward(
-                        input=latent_space_subsample,
-                        hidden=decoder_hidden[:,indx])
-                pyro.sample("aa_sequences", dist.Categorical(logits=logits), obs=aminoacid_sequences[indx]) #aa_seq = [n_nodes,align_seq_len]
+        blosum = self.blosum_weighted.repeat(latent_space.shape[0],1).reshape(latent_space.shape[0],self.align_seq_len,self.aa_probs) #[n_nodes,max_seq,21] #Highlight: it workedwith the entire blosum weighted matrix
+        #blosum = batch_blosum.repeat(latent_space.shape[0],1).reshape(latent_space.shape[0],self.max_seq_len,self.aa_prob) #[n_nodes,max_seq,21] #only use the weighted average of the batch sequences
+        blosum = self.embed(blosum)
+        latent_space = torch.cat((latent_space,blosum),dim=2) #[n_nodes,max_seq_len,z_dim + 21]
+        decoder_hidden = self.h_0_MODEL.expand(self.decoder.num_layers * 2, latent_space.shape[0],
+                                               self.gru_hidden_dim).contiguous()  # bidirectional
+        with pyro.plate("plate_len", aminoacid_sequences.shape[1], dim=-1), pyro.plate("plate_seq",aminoacid_sequences.shape[0],dim=-2):
+            logits = self.decoder.forward(
+                    input=latent_space,
+                    hidden=decoder_hidden)
+            pyro.sample("aa_sequences", dist.Categorical(logits=logits), obs=aminoacid_sequences) #aa_seq = [n_nodes,max_seq_len]
 
     def sample(self, map_estimates, n_samples, family_data_test, patristic_matrix,cladistic_matrix,use_argmax=False,use_test=True,use_test2=False):
-        if use_test or use_test2:
+        """Samples using all sequences, which is not computationally feasible if there is a high number of sequences"""
+        if use_test2: #MAP estimate
+            assert patristic_matrix[1:,1:].shape == (self.n_all,self.n_all)
+            latent_space = self.conditional_samplingMAP(map_estimates,patristic_matrix)
+            n_nodes = self.n_internal #I had to split it up because of some weird data cases (coral), otherwise family_data_test.shape[0] would have sufficed
+        elif use_test:# Marginal posterior
             assert patristic_matrix[1:,1:].shape == (self.n_all,self.n_all)
             latent_space = self.conditional_sampling(map_estimates,patristic_matrix)
             n_nodes = self.n_internal #I had to split it up because of some weird data cases (coral), otherwise family_data_test.shape[0] would have sufficed
-        else:
-            latent_space = map_estimates["latent_z"].T
-            assert latent_space.shape == (self.n_leaves, self.z_dim)
+        else:# sample training leaves
             n_nodes = self.n_leaves
+            latent_space = map_estimates["latent_z"].T
+            assert latent_space.shape == (n_nodes, self.z_dim)
 
         decoder_hidden = self.h_0_MODEL.expand(self.decoder.num_layers * 2, latent_space.shape[0],self.gru_hidden_dim).contiguous()  # Not bidirectional
-        latent_space_extended = latent_space.repeat(1, self.align_seq_len).reshape(n_nodes,self.align_seq_len, self.z_dim)
-        blosum = self.blosum_weighted.repeat(latent_space_extended.shape[0], 1).reshape(latent_space_extended.shape[0], self.align_seq_len,self.aa_probs)  # [n_nodes,max_seq,21]
+        latent_space_ = latent_space.repeat(1, self.max_seq_len).reshape(n_nodes,self.align_seq_len, self.z_dim)
+        blosum = self.blosum_weighted.repeat(latent_space_.shape[0], 1).reshape(latent_space_.shape[0], self.align_seq_len,self.aa_probs)  # [n_nodes,max_seq,21]
         blosum = self.embed(blosum)
-        latent_space_extended = torch.cat((latent_space_extended, blosum), dim=2)  # [n_nodes,align_seq_len,z_dim + 21]
+        latent_space_ = torch.cat((latent_space_, blosum), dim=2)  # [n_nodes,max_seq_len,z_dim + 21]
 
-        with pyro.plate("plate_len",self.align_seq_len, dim=-1):
-            with pyro.plate("plate_seq",n_nodes,dim=-2,subsample_size=n_nodes):
-                logits = self.decoder.forward(
-                    input=latent_space_extended,
-                    hidden=decoder_hidden)
-                if use_argmax:#Pick the sequence with the highest likelihood, now n_samples, n_samples = 1
-                    aa_sequences = torch.argmax(logits,dim=2).unsqueeze(0) #I add one dimension at the beginning to resemble 1 sample and not have to change all the plotting code
-                else:
-                    aa_sequences = dist.Categorical(logits=logits).sample([n_samples])
+        with pyro.plate("plate_len",self.max_seq_len, dim=-1), pyro.plate("plate_seq",n_nodes,dim=-2):
+            logits = self.decoder.forward(
+                input=latent_space_,
+                hidden=decoder_hidden)
+            if use_argmax:
+                #Pick the sequence with the highest likelihood, now n_samples, n_samples = 1
+                aa_sequences = torch.argmax(logits,dim=2).unsqueeze(0) #I add one dimension at the beginning to resemble 1 sample and not have to change all the plotting code
+            else:
+                aa_sequences = dist.Categorical(logits=logits).sample([n_samples])
+        #return aa_sequences,latent_space, logits, None, None
         sampling_out = SamplingOutput(aa_sequences=aa_sequences.detach(),
                                       latent_space=latent_space.detach(),
                                       logits=logits.detach(),
@@ -545,37 +593,83 @@ class DRAUPNIRModel_plating(DRAUPNIRModelClass):
 
         return sampling_out
 
+    def sample_batched(self, map_estimates, n_samples, family_data_test, patristic_matrix_full,patristic_matrix_test,batch_idx=None,use_argmax=False,use_test=True,use_test2=False):
+        """Batched sampling for large data sets"""
+
+        if use_test or use_test2:# Only Marginal posterior available when batching
+            assert patristic_matrix_full[1:,1:].shape == (self.n_all,self.n_all)
+            #Highlight: Slice out the train sequences and only a batch from the test sequences
+            if batch_idx[1] is None:
+                self.internal_nodes_batch = patristic_matrix_test[int(batch_idx[0]) + 1:, 0]
+            else:
+                self.internal_nodes_batch = patristic_matrix_test[int(batch_idx[0])+1:int(batch_idx[1])+1,0]
+            self.n_internal_batch = len(self.internal_nodes_batch)
+            nodes_batch = torch.cat((self.leaves_nodes,self.internal_nodes_batch))
+            self.n_leaves_internal_batch = len(nodes_batch)
+            indexes = (patristic_matrix_full[:, 0][..., None] == nodes_batch).any(-1)
+            indexes[0] = True #re-add the nodes names
+            patristic_matrix = patristic_matrix_full[indexes]
+            patristic_matrix = patristic_matrix[:,indexes]
+            latent_space = self.conditional_sampling_batch(map_estimates,patristic_matrix)
+            n_nodes = self.n_internal_batch
+        else: #training/leaves
+            n_nodes = self.n_leaves_batch #here n_leaves has been overloaded by the batch size
+            latent_space = map_estimates["latent_z"].T
+            latent_space = latent_space[int(batch_idx[0]):int(batch_idx[1])]
+            assert latent_space.shape == (n_nodes, self.z_dim)
+
+        decoder_hidden = self.h_0_MODEL.expand(self.decoder.num_layers * 2, latent_space.shape[0],self.gru_hidden_dim).contiguous()  # Not bidirectional
+        latent_space_ = latent_space.repeat(1, self.align_seq_len).reshape(n_nodes,self.align_seq_len, self.z_dim)
+        blosum = self.blosum_weighted.repeat(latent_space_.shape[0], 1).reshape(latent_space_.shape[0], self.align_seq_len,self.aa_probs)  # [n_nodes,max_seq,21]
+        blosum = self.embed(blosum)
+        latent_space_ = torch.cat((latent_space_, blosum), dim=2)  # [n_nodes,max_seq_len,z_dim + 21]
+
+        with pyro.plate("plate_len",self.align_seq_len, dim=-1), pyro.plate("plate_seq",n_nodes,dim=-2):
+            logits = self.decoder.forward(
+                input=latent_space_,
+                hidden=decoder_hidden)
+            if use_argmax:
+                #Pick the sequence with the highest likelihood, now n_samples, n_samples = 1
+                aa_sequences = torch.argmax(logits,dim=2).unsqueeze(0) #I add one dimension at the beginning to resemble 1 sample and not have to change all the plotting code
+            else:
+                aa_sequences = dist.Categorical(logits=logits).sample([n_samples])
+        sampling_out = SamplingOutput(aa_sequences=aa_sequences.detach(),
+                                      latent_space=latent_space.detach(),
+                                      logits=logits.detach(),
+                                      phis=None,
+                                      psis=None,
+                                      mean_phi=None,
+                                      mean_psi=None,
+                                      kappa_phi=None,
+                                      kappa_psi=None)
+
+        return sampling_out
 
 class DRAUPNIRModel_cladebatching(DRAUPNIRModelClass):
-    """Implements the clade batched version of Draupnir with full latent space inference.
-     a) It receives as an input a clade of the leaves dataset
-     b) uses the unsplitted leaves patristic matrix for full latent space inference
-     c} uses a GRU as the mapping function and blosum embeddings."""
+    """Perform inference by dividing the tree into batches that correspond to the clade in the treeClade batch training and sampling, with full latent space inference"""
     def __init__(self,ModelLoad):
         DRAUPNIRModelClass.__init__(self,ModelLoad)
         self.rnn_input_size = self.z_dim + self.aa_probs
-        #self.decoder_attention = RNNAttentionDecoder(self.n_leaves, self.align_seq_len, self.aa_probs, self.gru_hidden_dim,self.rnn_input_size,self.embedding_dim, self.z_dim, self.kappa_addition)
+        #self.decoder_attention = RNNAttentionDecoder(self.n_leaves, self.max_seq_len, self.aa_prob, self.gru_hidden_dim,self.rnn_input_size,self.embedding_dim, self.z_dim, self.kappa_addition)
         self.decoder = RNNDecoder_Tiling(self.align_seq_len, self.aa_probs, self.gru_hidden_dim, self.z_dim, self.rnn_input_size,self.kappa_addition,self.num_layers,self.pretrained_params)
         self.embed = EmbedComplex(self.aa_probs,self.embedding_dim, self.pretrained_params)
-    def model(self, family_data, patristic_matrix_sorted,cladistic_matrix,clade_blosum):
+    def model(self, family_data, patristic_matrix_sorted,cladistic_matrix,data_blosum,batch_blosum):
         aminoacid_sequences = family_data[:, 2:, 0]
-        batch_nodes = family_data[:, 0, 1]
-        batch_indexes = (patristic_matrix_sorted[1:, 0][..., None] == batch_nodes).any(-1)
+        #batch_nodes = family_data[:, 0, 1]
+        #batch_indexes = (patristic_matrix_sorted[1:, 0][..., None] == batch_nodes).any(-1)
         # Highlight: Register GRU module
         pyro.module("embeddings",self.embed)
-        #pyro.module("decoder_attention", self.decoder_attention)
         pyro.module("decoder", self.decoder)
+        self.n_leaves_batch = family_data.shape[0]
         # Highlight: GP prior over the latent space of all the leaves
-        latent_space = self.gp_prior(patristic_matrix_sorted)
+        latent_space = self.gp_prior_batched(patristic_matrix_sorted) #TODO: make a function model and one model_batched?
         # Highlight: MAP the latent space to logits using the Decoder from a Seq2seq model with/without attention
         latent_space = latent_space.repeat(1,self.align_seq_len).reshape(latent_space.shape[0],self.align_seq_len,self.z_dim) #[n_nodes,max_seq,z_dim]
-        #if clade_blosum is not None: #Highlight: For clade training we only use the blosum information of that part of the alignment that includes the sequences in the clade TODO: Move outside training
-        blosum = clade_blosum.repeat(latent_space.shape[0],1).reshape(latent_space.shape[0],self.align_seq_len,self.aa_probs) #[n_nodes,max_seq,21]
-        #else: #blosum embedding is based on the entire alignment
-        #blosum = self.blosum_weighted.repeat(latent_space.shape[0],1).reshape(latent_space.shape[0],self.align_seq_len,self.aa_probs) #[n_nodes,max_seq,21]
+        #blosum = batch_blosum.repeat(latent_space.shape[0],1).reshape(latent_space.shape[0],self.max_seq_len,self.aa_prob) #[n_nodes,max_seq,21]
+        blosum = self.blosum_weighted.repeat(latent_space.shape[0],1).reshape(latent_space.shape[0],self.align_seq_len,self.aa_probs) #[n_nodes,max_seq,21]
         blosum = self.embed(blosum)
-        latent_space = torch.cat((latent_space,blosum),dim=2) #[n_nodes,align_seq_len,z_dim + 21]
-        batch_latent_space = latent_space[batch_indexes] #In order to reduce the load on the GRU memory we split the latent space of the leaves by clades/batch
+        batch_latent_space = torch.cat((latent_space,blosum),dim=2) #[n_nodes,max_seq_len,z_dim + 21]
+        #batch_latent_space = latent_space[batch_indexes] #Highlight: For plating; In order to reduce the load on the GRU memory we split the latent space of the leaves by clades/batch
         decoder_hidden = self.h_0_MODEL.expand(self.decoder.num_layers * 2,
                                                batch_latent_space.shape[0],
                                                self.gru_hidden_dim).contiguous()  # bidirectional
@@ -583,9 +677,10 @@ class DRAUPNIRModel_cladebatching(DRAUPNIRModelClass):
             logits = self.decoder.forward(
                     input=batch_latent_space,
                     hidden=decoder_hidden)
-            pyro.sample("aa_sequences", dist.Categorical(logits=logits), obs=aminoacid_sequences) #aa_seq = [n_nodes,align_seq_len]
+            pyro.sample("aa_sequences", dist.Categorical(logits=logits), obs=aminoacid_sequences) #aa_seq = [n_nodes,max_seq_len]
 
     def sample(self, map_estimates, n_samples, family_data_test, patristic_matrix,cladistic_matrix,use_argmax=False,use_test=True,use_test2=False):
+        """Full latent space inference, plating sequences by clade"""
         if use_test:
             assert patristic_matrix[1:,1:].shape == (self.n_all,self.n_all)
             latent_space = self.conditional_sampling(map_estimates,patristic_matrix)
@@ -603,7 +698,7 @@ class DRAUPNIRModel_cladebatching(DRAUPNIRModelClass):
         latent_space_extended = latent_space.repeat(1, self.align_seq_len).reshape(n_nodes,self.align_seq_len, self.z_dim)
         blosum = self.blosum_weighted.repeat(latent_space_extended.shape[0], 1).reshape(latent_space_extended.shape[0], self.align_seq_len,self.aa_probs)  # [n_nodes,max_seq,21]
         blosum = self.embed(blosum)
-        latent_space_extended = torch.cat((latent_space_extended, blosum), dim=2)  # [n_nodes,align_seq_len,z_dim + 21]
+        latent_space_extended = torch.cat((latent_space_extended, blosum), dim=2)  # [n_nodes,max_seq_len,z_dim + 21]
 
         with pyro.plate("plate_len",self.align_seq_len, dim=-1), pyro.plate("plate_seq",n_nodes,dim=-2):
             logits = self.decoder.forward(
@@ -613,6 +708,57 @@ class DRAUPNIRModel_cladebatching(DRAUPNIRModelClass):
                 aa_sequences = torch.argmax(logits,dim=2).unsqueeze(0) #I add one dimension at the beginning to resemble 1 sample and not have to change all the plotting code
             else:
                 aa_sequences = dist.Categorical(logits=logits).sample([n_samples])
+        sampling_out = SamplingOutput(aa_sequences=aa_sequences.detach(),
+                                      latent_space=latent_space.detach(),
+                                      logits=logits.detach(),
+                                      phis=None,
+                                      psis=None,
+                                      mean_phi=None,
+                                      mean_psi=None,
+                                      kappa_phi=None,
+                                      kappa_psi=None)
+        return sampling_out
+    def sample_batched(self, map_estimates, n_samples, family_data_test, patristic_matrix_full,patristic_matrix_test,batch_idx=None,use_argmax=False,use_test=True,use_test2=False):
+        """Batched sampling based on clade membership"""
+
+        if use_test or use_test2:# Only Marginal posterior available when batching
+            assert patristic_matrix_full[1:,1:].shape == (self.n_all,self.n_all)
+            #Highlight: Slice out the train sequences and only a batch from the test sequences
+            self.internal_nodes_batch = torch.tensor(batch_idx)
+            self.n_internal_batch = len(self.internal_nodes_batch)
+            nodes_batch = torch.cat((self.leaves_nodes,self.internal_nodes_batch))
+            self.n_leaves_internal_batch = len(nodes_batch)
+            indexes = (patristic_matrix_full[:, 0][..., None] == nodes_batch).any(-1)
+            indexes[0] = True #re-add the nodes names
+            patristic_matrix = patristic_matrix_full[indexes]
+            patristic_matrix = patristic_matrix[:,indexes]
+            latent_space = self.conditional_sampling_batch(map_estimates,patristic_matrix)
+            n_nodes = self.n_internal_batch
+        else: #training/leaves
+            #Highlight: In clade-batching nodes_batch = batch_idx !
+            n_nodes = len(batch_idx)
+            latent_space = map_estimates["latent_z"].T
+            indexes = (self.leaves_nodes[..., None] == torch.tensor(batch_idx)).any(-1)
+            #indexes = torch.eq(self.leaves_nodes, torch.tensor(batch_idx))
+            #Highlight: pick the latent space correspondant to the nodes in the clade
+            latent_space = latent_space[indexes]
+            assert latent_space.shape == (n_nodes, self.z_dim)
+
+        decoder_hidden = self.h_0_MODEL.expand(self.decoder.num_layers * 2, latent_space.shape[0],self.gru_hidden_dim).contiguous()  # Not bidirectional
+        latent_space_ = latent_space.repeat(1, self.align_seq_len).reshape(n_nodes,self.align_seq_len, self.z_dim)
+        blosum = self.blosum_weighted.repeat(latent_space_.shape[0], 1).reshape(latent_space_.shape[0], self.align_seq_len,self.aa_probs)  # [n_nodes,max_seq,21]
+        blosum = self.embed(blosum)
+        latent_space_ = torch.cat((latent_space_, blosum), dim=2)  # [n_nodes,max_seq_len,z_dim + 21]
+
+        with pyro.plate("plate_len",self.align_seq_len, dim=-1), pyro.plate("plate_seq",n_nodes,dim=-2):
+            logits = self.decoder.forward(
+                input=latent_space_,
+                hidden=decoder_hidden)
+            if use_argmax:
+                #Pick the sequence with the highest likelihood, now n_samples, n_samples = 1
+                aa_sequences = torch.argmax(logits,dim=2).unsqueeze(0) #I add one dimension at the beginning to resemble 1 sample and not have to change all the plotting code
+            else:
+                aa_sequences = dist.Categorical(logits=logits).sample([n_samples])
         #return aa_sequences,latent_space, logits, None, None
         sampling_out = SamplingOutput(aa_sequences=aa_sequences.detach(),
                                       latent_space=latent_space.detach(),
@@ -623,6 +769,7 @@ class DRAUPNIRModel_cladebatching(DRAUPNIRModelClass):
                                       mean_psi=None,
                                       kappa_phi=None,
                                       kappa_psi=None)
+
         return sampling_out
 
 class DRAUPNIRModel_leaftesting(DRAUPNIRModelClass):
@@ -633,7 +780,7 @@ class DRAUPNIRModel_leaftesting(DRAUPNIRModelClass):
         #self.decoder_attention = RNNAttentionDecoder(self.n_leaves, self.align_seq_len, self.aa_probs, self.gru_hidden_dim,self.rnn_input_size,self.embedding_dim, self.z_dim, self.kappa_addition)
         self.decoder = RNNDecoder_Tiling(self.align_seq_len, self.aa_probs, self.gru_hidden_dim, self.z_dim, self.rnn_input_size,self.kappa_addition,self.num_layers,self.pretrained_params)
         self.embed = EmbedComplex(self.aa_probs,self.embedding_dim, self.pretrained_params)
-    def model(self, family_data, patristic_matrix_sorted,cladistic_matrix,clade_blosum):
+    def model(self, family_data, patristic_matrix_sorted,cladistic_matrix,data_blosum,batch_blosum=None):
         aminoacid_sequences = family_data[:, 2:, 0]
         #angles = family_data[:, 2:, 1:3]
         train_nodes = family_data[:, 0, 1]
@@ -715,7 +862,7 @@ class DRAUPNIRModel_anglespredictions(DRAUPNIRModelClass):
         self.rnn_input_size = self.z_dim + self.aa_probs
         self.decoder = RNNDecoder_Tiling_Angles(self.align_seq_len, self.aa_probs, self.gru_hidden_dim, self.z_dim, self.rnn_input_size,self.kappa_addition,self.num_layers,self.pretrained_params)
         self.embed = EmbedComplex(self.aa_probs,self.embedding_dim, self.pretrained_params)
-    def model(self, family_data, patristic_matrix_sorted,cladistic_matrix,clade_blosum):
+    def model(self, family_data, patristic_matrix_sorted,cladistic_matrix,data_blosum,batch_blosum=None):
         aminoacid_sequences = family_data[:, 2:, 0]
         angles = family_data[:, 2:, 1:3]
         angles_mask = torch.where(angles == 0., angles, 1.).type(angles.dtype) #keep as 0 the gaps and set to 1 where there is an observation
